/*-
 * Copyright (c) 2012 Dag-Erling Smørgrav
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */

#include <sys/endian.h>
#include <sys/tree.h>

#include <err.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

struct attachtime {
	char name[24];
	uint64_t delta;
};

// typedef struct at_entry at_entry;
struct at_entry  {
	RB_ENTRY(at_entry) at_rb;
	char name[24];
	FILE *f;
	unsigned long count;
};

RB_HEAD(at_entry_tree, at_entry) at_entry_head;

static inline int
at_entry_cmp(const struct at_entry *a, const struct at_entry *b)
{

	return (strcmp(a->name, b->name));
}

RB_PROTOTYPE_STATIC(at_entry_tree, at_entry, at_rb, at_entry_cmp);
RB_GENERATE_STATIC(at_entry_tree, at_entry, at_rb, at_entry_cmp);

static int opt_v;

static void
init_at_entry(void)
{

	RB_INIT(&at_entry_head);
}

static struct at_entry *
get_at_entry(const char *name)
{
	
	struct at_entry key, *e;

	strcpy(key.name, name);
	if ((e = RB_FIND(at_entry_tree, &at_entry_head, &key)) != NULL)
		return (e);
	if ((e = calloc(1, sizeof *e)) == NULL)
		err(1, "calloc()");
	strcpy(e->name, name);
	if ((e->f = fopen(name, "a")) == NULL)
		err(1, "%s", name);
	RB_INSERT(at_entry_tree, &at_entry_head, e);
	return (e);
}

static void
exit_at_entry(void)
{
	struct at_entry *e;

	while ((e = RB_ROOT(&at_entry_head)) != NULL) {
		RB_REMOVE(at_entry_tree, &at_entry_head, e);
		fclose(e->f);
		free(e);
	}
}

static void
read_attachtimes(const char *fn)
{
	struct attachtime at;
	struct at_entry *e;
	ssize_t rlen;
	int fd;

	if ((fd = open(fn, O_RDONLY)) < 0)
		err(1, "%s", fn);
	init_at_entry();
	for (;;) {
		if ((rlen = read(fd, &at, sizeof at)) < 0)
			err(1, "%s", fn);
		if (rlen == 0)
			break;
		if (rlen != sizeof at) {
			warnx("%s: short read", fn);
			break;
		}
		if (*at.name == '\0')
			continue;
		e = get_at_entry(at.name);
		fprintf(e->f, "%llu\n", (unsigned long long)be64toh(at.delta));
		++e->count;
	}
	exit_at_entry();
	close(fd);
}

static void
usage(void)
{

	fprintf(stderr, "attachtimes [-v] [file ...]\n");
	exit(1);
}

int
main(int argc, char *argv[])
{
	int opt;

	while ((opt = getopt(argc, argv, "v")) != -1)
		switch (opt) {
		case 'v':
			++opt_v;
			break;
		default:
			usage();
		}

	argc -= optind;
	argv += optind;

	if (argc != 0)
		read_attachtimes("/dev/stdout");
	else
		while (argc--)
			read_attachtimes(*argv++);
	exit(0);
}
